(***********************************************************************)
(*                                                                     *)
(*                           Htmlc                                     *)
(*                                                                     *)
(*                    Pierre Weis, INRIA Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 1997-2007,                                               *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  Distributed only by permission.                                    *)
(*                                                                     *)
(***********************************************************************)

(* $Id: debug.ml,v 1.14 2015/11/18 01:07:34 weis Exp $ *)

(** {3 Some simple generic tools for debugging} *)

(** The [Debug] module should ensure a simple invariant: a vertical box of
  level 0 is permanently opened for formatter [Format.err_formatter]. *)

let init, close =
  let is_open = ref false in
  (fun () ->
   if not !is_open then begin
     Format.eprintf "@[<v 0>";
     is_open := true;
   end),
  (fun () ->
   if !is_open then begin
     Format.eprintf "@]@.";
     is_open := false;
   end)
;;

let flush () =
  (* Flush the formatter. *)
  close ();
  (* Reensure the invariant. *)
  init ();
;;

let get_debug, set_debug =
  let debug = ref false in
  (fun () -> !debug),
  (fun () ->
   if !debug = false then init ();
   debug := true)
;;

(** A special [Debug.failwith] function to be able to trace the failure
  strings. *)
let failwith failure_string =
  if get_debug () then begin
    Format.eprintf "@ %s: raising Failure %S@ "
      (Filename.basename Sys.executable_name)
      failure_string;
    flush ();
  end;
  raise (Failure failure_string)
;;

let debug fmt =
  if get_debug ()
  then Format.eprintf (fmt ^^ "@ ")
  else Format.ifprintf Format.std_formatter fmt
;;

let fdebug f = if get_debug () then f Format.err_formatter;;

let warning s =
  Format.eprintf "@ %s: warning, %s@ "
    (Filename.basename Sys.executable_name)
    s
;;

let message s =
  if get_debug () then begin
    Format.eprintf "@ %s@ " s;
    flush ();
  end
;;

let pause name =
  if get_debug () then begin
    Format.eprintf
      "@ [ Debug.pause: <%s>, press return to go on. ]@ "
      name;
    flush ();
    ignore (input_line stdin);
  end
;;

(*
 Local Variables:
  compile-command: "make"
  End:
*)
