(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: unumber_command_option.ml,v 1.14 2017/11/07 20:57:39 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module S = Ustring;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;

(* Shortcuts for the Basement library. *)

open Unumber_command_option_types;;

(* Number command option operations. *)

let print_opt_doc ppf = function
  | None -> ()
  | Some msg -> Format.fprintf ppf "@[%s@]@ " msg
;;

let make_doc arg_name full_name opt_doc skind name print_default =
  let ppf = Format.str_formatter in
  Format.fprintf ppf
    "@[<v 8><%s>@ \
    @[Set %s to <%s>.@]@ \
    %a\
    @[%s is %s.@]@ \
    @[Default %s is %t.@]\
    @]"
    arg_name
    full_name arg_name
    print_opt_doc opt_doc
    (S.capitalize name) skind
    name print_default;
  Format.flush_str_formatter ()
;;

module Integer_Make =
  functor (S : Unumber_command_option_spec.STRUCTURE with type t = int) ->
struct
  open S;;

  type t = S.t;;

  let print ppf = Format.fprintf ppf "%d";;

  let { Ar.get = get; set = set; } = Ar.make_aref_single name default;;

  let arg_name =
    match kind with
    | Regular | Range _ -> "int"
    | Nonnegative | Positive -> "nat"
  ;;
  let skind =
    match kind with
    | Regular -> "an integer"
    | Nonnegative -> "a natural number"
    | Positive -> "a positive integer"
    | Range (s1, s2) -> Printf.sprintf "an integer between %s and %s" s1 s2
  ;;
  let print_default ppf = print ppf (get ());;
  let doc = make_doc arg_name full_name opt_doc skind name print_default;;

  let lok = Ok.long_option_key_of_string name
  and sok_st =
    match opt_short_option_key with
    | None -> Ok.Automatic
    | Some sok -> Ok.Short (Ok.of_string sok)
  ;;

  let rco =
    {
     Co.key = lok;
     spec = Arg.Int set;
     doc = doc;
    }
  ;;

  let rco_with_sok = (sok_st, rco);;
end
;;

module Float_Make =
  functor (S : Unumber_command_option_spec.STRUCTURE with type t = float) ->
struct
  open S;;

  type t = S.t;;

  let print ppf = Format.fprintf ppf "%g";;

  let { Ar.get = get; set = set; } = Ar.make_aref_single name default;;

  let arg_name = "float";;
  let skind =
    match kind with
    | Regular -> "a floating-point number"
    | Nonnegative -> "a nonnegative floating-point number"
    | Positive -> "a positive floating-point number"
    | Range (s1, s2) ->
        Printf.sprintf"a floating-point number between %s and %s" s1 s2
  ;;
  let print_default ppf = print ppf (get ());;
  let doc = make_doc arg_name full_name opt_doc skind name print_default;;

  let lok = Ok.long_option_key_of_string name
  and sok_st =
    match opt_short_option_key with
    | None -> Ok.Automatic
    | Some sok -> Ok.Short (Ok.of_string sok)
  ;;

  let rco =
    {
     Co.key = lok;
     spec = Arg.Float set;
     doc = doc;
    }
  ;;

  let rco_with_sok = (sok_st, rco);;
end
;;
