(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: umatrix_nonempty_set.ml,v 1.22 2017/11/07 20:57:40 fclement Exp $ *)

(* Shortcuts for the Ulib library.*)

module Nat = Unatural;;
module P = Upervasives;;
module L1 = Ulist1;;
module A = Uarray;;
(*module Idx = Uindex;;*)

let _0_ = Nat._0_;;

(* Shortcuts for the Basement library.*)

open Misc_sprint;;

let printer_nat = Unatural_print.printer_t;;
let printer_nat_couple = printer_Couple printer_nat printer_nat;;
let printer_uindex = Uindex_print.printer_t printer_nat_couple;;
let sprint_uindex_list = sprint_list printer_uindex;;

(* Shortcuts for the Mapping library.*)

module Iidx = Uindex_index;;
module Iuniv = Uindex_universe;;
module Midx = Umatrix_index;;
module Muniv = Umatrix_universe;;

(* Matrix slot nonempty set operations. *)

let qual = P.qualify "Umatrix_nonempty_set";;
let invalid_arg = P.invalid_arg_qual qual;;
let failwith = P.failwith_qual qual;;

type size = Muniv.size;;
type element = Muniv.element;;     (* = Nat.t * Nat.t *)
type uindex = Muniv.index;;        (* = element Iuniv.element
                                      = (Nat.t * Nat.t) Idx.t *)
type index = element Iuniv.index;; (* = (Nat.t * Nat.t) Idx.t Idx.t *)

type t =
  {
   universe : Muniv.t;
   uindices : element Iuniv.t;
  }
;;

let same = ( == );;

(* Construction functions. *)

let full mu =
  let max_uidx = Midx.of_nat (Muniv.cardinal mu) in
  let uindices = Iuniv.of_nonempty_range Midx._0_ max_uidx in
  { universe = mu; uindices = uindices; }
;;

let make sz = full (Muniv.make sz);;

let check_valid_uindices fun_name mu uindices =
  let card = Muniv.cardinal mu in
  let invalid_index uidx = (Midx.to_nat uidx) >= card in
  let exists_invalid_index = L1.exists invalid_index uindices in
  if exists_invalid_index then
    let invalid_indices = L1.find_all invalid_index uindices in
    invalid_arg fun_name
      (Printf.sprintf
        "uindices are not all valid (i.e. < %s): %s"
        (sprint_nat "card" card)
        (sprint_uindex_list "invalid_indices" invalid_indices))
;;

let unsafe_of_list1 mu uindices =
  { universe = mu; uindices = Iuniv.of_list1 uindices; }
;;

let of_list1 mu uindices =
  check_valid_uindices "of_list1" mu uindices;
  unsafe_of_list1 mu uindices
;;

(* Access functions. *)

let cardinal mns = Iuniv.cardinal mns.uindices;;

let uindex mns = Iuniv.element mns.uindices;;

let choice mns =
  let idx0 = Iidx._0_ in
  let uidx0 = uindex mns idx0 in
  (idx0, uidx0)
;;

let element mns idx =
  let uidx = uindex mns idx in
  Muniv.element mns.universe uidx
;;

let same_universe mns1 mns2 = mns1.universe == mns2.universe;;

let check_same_universe fun_name mns1 mns2 =
  if not (same_universe mns1 mns2) then
    invalid_arg fun_name "universe clash"
;;

let equal mns1 mns2 =
  check_same_universe "equal" mns1 mns2;
  same mns1 mns2 ||
  Iuniv.equal mns1.uindices mns2.uindices
;;

let array_set a x _idx uidx = A.set a (Midx.to_nat uidx) x;;

(* Iterators. *)

let iter f mns = Iuniv.iter f mns.uindices;;

let sub_iter f mns = Iuniv.sub_iter f mns.uindices;;

let fold_left f accu mns = Iuniv.fold_left f accu mns.uindices;;

let full_fold_left f accu ({ universe = mu; _; } as mns) =
  let len = cardinal mns in
  let (_, nj) = Muniv.size mu in
  let rec loop accu idx uidx (i, j as x) =
    if Iidx.to_nat idx = len then accu else
    let accu = f accu idx uidx x in
    let idx = Iidx.succ idx in
    let uidx = Midx.succ uidx in
    let (i, j) =
      if j = Nat.pred nj then (Nat.succ i, _0_) else (i, Nat.succ j) in
    loop accu idx uidx (i, j) in
  loop accu Iidx._0_ Midx._0_ (_0_, _0_)
;;

let full_iter f = full_fold_left (fun _ idx uidx x -> f idx uidx x) ();;

(* Nonempty subsets. *)

type nonempty_subset =
  {
   support : t;
   indices : index array;
   nonempty_set : t;
  }
;;

let is_full mns = cardinal mns = Muniv.cardinal mns.universe;;

let check_full fun_name mns =
  if not (is_full mns) then
    invalid_arg fun_name "nonempty set is not full"
;;

let check_same_full fun_name mns1 mns2 =
  check_same_universe fun_name mns1 mns2;
  check_full fun_name mns1;
  check_full fun_name mns2
;;

let (full_index_of_uindex : uindex -> index) =
  fun uidx -> Iidx.of_nat (Midx.to_nat uidx)
;;

(* We use here the fact that the full set, containing all the elements of the
  universe, is numbered in the same order as the universe.
  Hence, the use of magic function full_index_of_uindex.
  Should efficiency be an issue, consider using Obj.magic on the whole array of
  indices. *)
let subset_of_full full mns =
  check_full "subset_of_full" full;
  check_same_universe "subset_of_full" full mns;
  let indices =
    A.map full_index_of_uindex (Iuniv.elements mns.uindices) in
  { support = full; indices = indices; nonempty_set = mns; }
;;

type dichotomy = nonempty_subset * nonempty_subset;;

let dichotomy_aux p mns =
  let dicho (tl, fl) idx uidx =
    if p idx uidx then (idx :: tl, fl) else (tl, idx :: fl) in
  let (tl, fl) =
    match fold_left dicho ([], []) mns with
    | ([], _) | (_, []) -> failwith "dichotomy" "empty subset"
    | (tx :: tl, fx :: fl) -> ((tx, tl), (fx, fl)) in
  let utl = L1.rev_map (uindex mns) tl
  and ufl = L1.rev_map (uindex mns) fl in
  let tmns = unsafe_of_list1 mns.universe utl
  and fmns = unsafe_of_list1 mns.universe ufl in
  ({ support = mns; indices = A.of_list1 tl; nonempty_set = tmns; },
   { support = mns; indices = A.of_list1 fl; nonempty_set = fmns; })
;;

let dichotomy p mns =
  try Some (dichotomy_aux p mns) with
  | Failure _  -> None
;;
