(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: configuration.ml,v 1.34 2017/11/13 12:40:10 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L1 = Ulist1;;
module S = Ustring;;
module Ar = Uaref;;
module Dbg = Udebug;;

(* Shortcuts for the Refinement library. *)

module Opt = Options;;

open Refinement_types;;
open Options_types;;
open Options;;
open Configuration_types;;

(* Configuration operations. *)

let qual = P.qualify "Configuration";;
let failwith = P.failwith_qual qual;;
let invalid_arg = P.invalid_arg_qual qual "";;

type t = Configuration_types.t;;

(* Uhistory.Make functor interface. *)

let printer = Configuration_types_print.printer_t;;
let print = Configuration_types_print.print_t;;

let get_external_program_command () =
  match Opt.get_option_external_program_command () with
  | [] -> invalid_arg "empty external program command path"
  | _cmd :: _args as sl -> S.concat " " sl
;;

let get_dimension () = Nat.of_int (get_option_dimension ());;

let get_dimension_management = get_option_dimension_management;;

let get_refinement_strategy () =
  match get_option_refinement_strategy () with
  | Option_Best_in_a_family ->
      let nct = Nat.of_int (get_option_number_of_cuts ()) in
      let bcfl = get_option_basic_refinement_families () in
      Best_in_a_family (nct, L1.of_list bcfl)
  | Option_Overall_best -> Overall_best
;;

let get_candidate_refinements_selection_criterion () =
  let r = get_option_candidate_refinements_selection_criterion_ratio ()
  and n =
    get_option_candidate_refinements_selection_criterion_maximum_number () in
  if r < 0. || r > 1. then
    failwith "get_candidate_refinements_selection_criterion"
      "ratio is outside [0,1]" else
  if n <= 0 then
    failwith "get_candidate_refinements_selection_criterion"
      "maximum number is nonpositive" else
  (r, Nat.of_int n)
;;

let get_stopping_criterion_config () =
  {
   min_relative_gradients_norm =
     get_option_min_relative_gradients_norm ();
   min_relative_best_scalar_indicator =
     get_option_min_relative_best_scalar_indicator ();
   min_relative_scalar_cost_decrease =
     get_option_min_relative_scalar_cost_decrease ();
   max_iteration = Nat.of_int (get_option_max_iteration ());
  }
;;

let { Ar.get = get_configuration_opt; set = set_configuration_opt; } =
  Ar.make_aref_unique "algorithm configuration" None
;;

let cold_set () =
  let config =
    {
     external_program_command = get_external_program_command ();
     dimension = get_dimension ();
     dimension_management = get_dimension_management ();
     refinement_strategy = get_refinement_strategy ();
     candidate_refinements_selection_criterion =
       get_candidate_refinements_selection_criterion ();
     stopping_criterion_config = get_stopping_criterion_config ();
    } in
  set_configuration_opt (Some config)
;;

let warm_set config =
  let config =
    { config with stopping_criterion_config = get_stopping_criterion_config (); } in
  set_configuration_opt (Some config)
;;

let get () =
  match get_configuration_opt () with
  | None -> assert false
  | Some config -> config
;;

(* Algorithm configuration. *)

let phase_dimension = Opt.phase_dimension;;

let get_external_program_command () = (get ()).external_program_command;;
let get_dimension () = (get ()).dimension;;
let get_dimension_management () = (get ()).dimension_management;;
let get_refinement_strategy () = (get ()).refinement_strategy;;
let get_candidate_refinements_selection_criterion () =
  (get ()).candidate_refinements_selection_criterion;;
let get_stopping_criterion_config () = (get ()).stopping_criterion_config;;

(* Program configuration. *)

let get_formatter () =
  match get_option_out_formatter () with
  | Option_Std_out -> Format.std_formatter
  | Option_Std_err -> Format.err_formatter
  | Option_Out_file fname ->
      Format.formatter_of_out_channel (Pervasives.open_out fname)
;;

let get_report_allocated_memory = get_option_report_allocated_memory;;

let get_incremental_mode = get_option_incremental_mode;;

let get_graphics_mode = get_option_graphics_mode;;

let check_graphics_window_sizes () =
  let w_min = get_option_graphics_window_min_width ()
  and h_min = get_option_graphics_window_min_height () in
  let w_max = get_option_graphics_window_max_width ()
  and h_max = get_option_graphics_window_max_height () in
  if w_max < w_min then
    invalid_arg "Min width is greater than max width" else
  if h_max < h_min then
    invalid_arg "Min height is greater then max height"
;;

let get_graphics_window_min_size () =
  check_graphics_window_sizes ();
  let w_min = get_option_graphics_window_min_width ()
  and h_min = get_option_graphics_window_min_height () in
  (Nat.of_int w_min, Nat.of_int h_min)
;;

let get_graphics_window_max_size () =
  check_graphics_window_sizes ();
  let w_max = get_option_graphics_window_max_width ()
  and h_max = get_option_graphics_window_max_height () in
  (Nat.of_int w_max, Nat.of_int h_max)
;;

let check_color c =
  if c < 0 || c > 0xFFFFFF then
    invalid_arg "color is not in the range [0..0xFFFFFF]"
;;

let get_graphics_window_bg_color () =
  let bg = get_option_graphics_window_background_color () in
  check_color bg;
  bg
;;

let num_steps = 256;;

let make_color_mapping = function
  | Gray -> Colormapping.make_rgb Graphics.black Graphics.white
  | Cool -> Colormapping.make_rgb Graphics.cyan Graphics.magenta
  | Hot -> Colormapping.make_hsb Graphics.red Graphics.yellow
  | Jet -> Colormapping.make_hsb Graphics.red Graphics.blue
;;

let get_graphics_parameter_color_mapping () =
  let csn = get_option_graphics_parameter_color_scale_name () in
  let min = get_option_graphics_parameter_min ()
  and max = get_option_graphics_parameter_max () in
  if max < min then
    failwith "get_graphics_parameter_color_mapping"
      "min parameter is greater than max parameter" else
  make_color_mapping csn num_steps min max
;;

let get_graphics_gradient_color_mapping () =
  let csn = get_option_graphics_gradient_color_scale_name () in
  let max = get_option_graphics_gradient_amplitude () in
  if max < 0. then
    failwith "get_graphics_gradient_color_mapping"
      "gradient amplitude is negative" else
  let min = -. max in
  make_color_mapping csn num_steps min max
;;

(* Monitoring. *)

let {
  Dbg.get_warn_error = get_warn_error;
  report = report;
  debug = debug;
  warning = warning;
  error = error;
  report_action = report_action;
  debug_action = debug_action;
  warning_action = warning_action;
  error_action = error_action;
} = Opt.printing_facilities
;;

let debug fun_name = debug (Some fun_name)
and warning fun_name = warning (Some fun_name)
and error fun_name = error (Some fun_name)
and debug_action fun_name = debug_action (Some fun_name)
and warning_action fun_name = warning_action (Some fun_name)
and error_action fun_name = error_action (Some fun_name)
;;

let print_ident_value ident print ppf value =
  Format.fprintf ppf "%s =@ %a" ident print value
;;

let print_report print ppf value =
  report_action (fun () -> Format.fprintf ppf "@[<hv>%a@]@." print value; "")
;;

let print_debug fun_name ident print ppf value =
  debug_action fun_name
    (fun () ->
      Format.fprintf ppf "@[<v 2>%s:@ %a@]@."
        fun_name (print_ident_value ident print) value; "")
;;

let print_debug_no_arg fun_name print ppf =
  debug_action fun_name
    (fun () -> Format.fprintf ppf "@[<v 2>%s:@ %t@]@." fun_name print; "")
;;
