(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: options.ml,v 1.51.2.2 2017/11/26 14:37:50 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L = Ulist;;
module A = Uarray;;
module Ar = Uaref;;
module Ok = Uoption_key;;
module Co = Ucommand_option;;
module Dbg = Udebug;;
module Pic = Upph_id_card;;

let _0_ = Nat._0_;;

let string_of_list s_of l =
  let len = L.length l in
  L.fold_left_i
    (fun i accu x ->
      if i = _0_ then accu ^ s_of x else
      if i = Nat.pred len then accu ^ ", and " ^ s_of x else
      accu ^ ", " ^ s_of x)
    "" l
;;

(* Shortcuts for the Basement library. *)

module Nco = Unumber_command_option;;
module Sco = Ustring_command_option;;

open Unumber_command_option_types;;
open Refinement_types;;
open Options_types;;

(* Option operations. *)

(* Algorithm configuration. *)

let qual = P.qualify "Options";;
let failwith = P.failwith_qual qual;;
let invalid_arg = P.invalid_arg_qual qual;;

let { Ar.get_list = get_option_external_program_command;
      add_item = add_option_external_program_command_item; _; } =
  Ar.make_aref_list []
;;

(* External program command option. *)

let external_program_command_lok =
  Ok.long_option_key_of_string "external program command"
and external_program_command_sok_st =
  Ok.Short (Ok.of_string "--")
;;

let external_program_command_rco =
  {
   Co.key = external_program_command_lok;
   spec = Arg.Rest add_option_external_program_command_item;
   doc =
     Printf.sprintf
       "<command_line>\n\
        \tTreat the remaining arguments as the command line to launch\n\
        \tthe external worker.\n\
        \tThis must be the last arguments of the present command."
  }
;;

let external_program_command_rco_with_sok =
  (external_program_command_sok_st, external_program_command_rco)
;;

(* Dimension option. *)

module Dimension_structure =
struct
  type t = option_dimension;;

  let kind = Positive;;
  let name = "dimension";;
  let full_name = Printf.sprintf "%s of the parameter in each point" name;;
  let default = 1;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-dim";;
end
;;

module Dimension = Nco.Integer_Make (Dimension_structure);;
let ((_, { Co.key = dimension_lok; _; }) as dimension_rco_with_sok) =
  Dimension.rco_with_sok
and get_option_dimension = Dimension.get
;;

(* Dimension management option. *)

module Dimension_management_structure =
struct
  type t = option_dimension_management;;

  let name = "dimension management";;
  let plural_name = Printf.sprintf "%ss" name;;

  let default = Vector
  and all = [ Vector; Best_component_only; ]
  ;;

  let of_string = function
    | "vector" | "v" -> Vector
    | "best_component_only" | "best_only" | "bco" -> Best_component_only
    | s ->
        failwith "dimension_management_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;

  let to_string = function
    | Vector -> "vector"
    | Best_component_only -> "best_component_only"
  ;;

  let opt_short_option_key = None;;
end
;;

module Dimension_management = Sco.Make (Dimension_management_structure);;
let dimension_management_rco_with_sok = Dimension_management.rco_with_sok
and get_option_dimension_management = Dimension_management.get
;;

(* Refinement strategy option. *)

module Refinement_strategy_structure =
struct
  type t = option_refinement_strategy;;

  let name = "refinement strategy"
  and plural_name = "refinement strategies"
  ;;

  let default = Option_Overall_best
  and all = [ Option_Best_in_a_family; Option_Overall_best; ]
  ;;

  let of_string = function
    | "best_in_a_family" | "bf" | "family" | "f" -> Option_Best_in_a_family
    | "overall_best" | "ob" | "best" | "b" -> Option_Overall_best
    | s ->
        failwith "refinement_strategy_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;

  let to_string = function
    | Option_Best_in_a_family -> "best_in_a_family"
    | Option_Overall_best -> "overall_best"
  ;;

  let opt_short_option_key = None;;
end
;;

module Refinement_strategy = Sco.Make (Refinement_strategy_structure);;
let refinement_strategy_rco_with_sok = Refinement_strategy.rco_with_sok
and get_option_refinement_strategy = Refinement_strategy.get
;;

(* Number of cuts option. *)

module Number_of_cuts_structure =
struct
  type t = option_number_of_cuts;;

  let kind = Nonnegative;;
  let name = "number of cuts";;
  let full_name = Printf.sprintf "%s for each basic refinement family" name;;
  let default = 0;;
  let opt_doc =
    Some "By convention, 0 means all possible cuts of the family."
  ;;
  let opt_short_option_key = Some "-nc";;
end
;;

module Number_of_cuts = Nco.Integer_Make (Number_of_cuts_structure);;
let number_of_cuts_rco_with_sok = Number_of_cuts.rco_with_sok
and get_option_number_of_cuts = Number_of_cuts.get
;;

(* Add basic refinement family option. *)

let all_option_basic_refinement_families = [ Horizontal_cut; Vertical_cut; ];;

let default_option_basic_refinement_families =
  all_option_basic_refinement_families
;;

let { Ar.get_list = get_option_basic_refinement_families;
      clear_list = clear_option_basic_refinement_families;
      add_item = add_option_basic_refinement_family; } =
  Ar.make_aref_list default_option_basic_refinement_families
;;

let clear_basic_refinement_families_lok =
  Ok.long_option_key_of_string "clear basic refinement families"
and clear_basic_refinement_families_sok_st =
  Ok.Short (Ok.of_string "-clear-bcf")
;;

let clear_basic_refinement_families_rco =
  {
   Co.key = clear_basic_refinement_families_lok;
   spec = Arg.Unit clear_option_basic_refinement_families;
   doc = Printf.sprintf "\n\tClear the basic refinement family list.";
  }
;;

let clear_basic_refinement_families_rco_with_sok =
  (clear_basic_refinement_families_sok_st,
   clear_basic_refinement_families_rco)
;;

let option_basic_refinement_family_of_string = function
  | "horizontal" | "hor" | "h" -> Horizontal_cut
  | "vertical" | "ver" | "v" -> Vertical_cut
  | s ->
      failwith "basic_refinement_family_of_string"
        (Printf.sprintf "Unknown basic refinement family %S" s)
;;

let add_option_basic_refinement_family_from_string =
  P.compose
    add_option_basic_refinement_family
    option_basic_refinement_family_of_string
;;

let string_of_option_basic_refinement_family = function
  | Horizontal_cut -> "horizontal"
  | Vertical_cut -> "vertical"
;;

let print_all_option_basic_refinement_families () =
  string_of_list
    string_of_option_basic_refinement_family
    all_option_basic_refinement_families
;;

let print_default_option_basic_refinement_family_list () =
  string_of_list
    string_of_option_basic_refinement_family
    (get_option_basic_refinement_families ())
;;

let add_basic_refinement_family_lok =
  Ok.long_option_key_of_string "add basic refinement family"
and add_basic_refinement_family_sok_st =
  Ok.Short (Ok.of_string "-add-bcf")
;;

let add_basic_refinement_family_rco =
  {
   Co.key = add_basic_refinement_family_lok;
   spec = Arg.String add_option_basic_refinement_family_from_string;
   doc =
     Printf.sprintf
       "<str>\n\
        \tAdd <str> to the list of basic refinement families.\n\
        \tUsed when the refinement strategy is set to %S.\n\
        \tValid basic refinement families are: %t.\n\
        \tDefault list is %t."
       (Refinement_strategy_structure.to_string Option_Best_in_a_family)
       print_all_option_basic_refinement_families
       print_default_option_basic_refinement_family_list;
  }
;;

let add_basic_refinement_family_rco_with_sok =
  (add_basic_refinement_family_sok_st, add_basic_refinement_family_rco)
;;

(* Candidate refinements selection criterion options. *)

module Candidate_refinements_selection_criterion_ratio_structure =
struct
  type t = option_candidate_refinements_selection_criterion_ratio;;

  let kind = Range ("0", "1");;
  let name = "selection criterion ratio";;
  let full_name = Printf.sprintf "candidate refinement %s" name;;
  let default = 0.8;;
  let opt_doc =
    Some "Candidate refinements with indicator smaller than \
          the highest indicator\n\
          \ttimes <float> will not go through the optimization phase."
  ;;
  let opt_short_option_key = Some "-sc-ratio";;
end
;;

module Candidate_refinements_selection_criterion_maximum_number_structure =
struct
  type t = option_candidate_refinements_selection_criterion_maximum_number;;

  let kind = Positive;;
  let name = "selection criterion maximum number";;
  let full_name =
    Printf.sprintf "candidate refinement %s" name
  ;;
  let default = 4;;
  let opt_doc =
    Some "At most <nat> candidate refinements \
          will go through the optimization phase."
  ;;
  let opt_short_option_key = Some "-sc-max-num";;
end
;;

module Candidate_refinements_selection_criterion_ratio =
  Nco.Float_Make
    (Candidate_refinements_selection_criterion_ratio_structure);;
let candidate_refinements_selection_criterion_ratio_rco_with_sok =
  Candidate_refinements_selection_criterion_ratio.rco_with_sok
and get_option_candidate_refinements_selection_criterion_ratio =
  Candidate_refinements_selection_criterion_ratio.get
;;

module Candidate_refinements_selection_criterion_maximum_number =
  Nco.Integer_Make
    (Candidate_refinements_selection_criterion_maximum_number_structure);;
let candidate_refinements_selection_criterion_maximum_number_rco_with_sok =
  Candidate_refinements_selection_criterion_maximum_number.rco_with_sok
and get_option_candidate_refinements_selection_criterion_maximum_number =
  Candidate_refinements_selection_criterion_maximum_number.get
;;

(* Min relative gradients norm option. *)

module Min_relative_gradients_norm_structure =
struct
  type t = option_min_relative_gradients_norm;;

  let kind = Nonnegative;;
  let gname = "relative gradients norm";;
  let name = Printf.sprintf "min %s" gname;;
  let full_name = Printf.sprintf "minimum %s" gname;;
  let default = 1e-6;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-stop-grad";;
end
;;

module Min_relative_gradients_norm =
  Nco.Float_Make (Min_relative_gradients_norm_structure)
;;
let min_relative_gradients_norm_rco_with_sok =
  Min_relative_gradients_norm.rco_with_sok
and get_option_min_relative_gradients_norm =
  Min_relative_gradients_norm.get
;;

(* Min relative best scalar indicator option. *)

module Min_relative_best_scalar_indicator_structure =
struct
  type t = option_min_relative_best_scalar_indicator;;

  let kind = Nonnegative;;
  let gname = "relative best scalar indicator";;
  let name = Printf.sprintf "min %s" gname;;
  let full_name = Printf.sprintf "minimum %s" gname;;
  let default = 1e-6;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-stop-indic";;
end
;;

module Min_relative_best_scalar_indicator =
  Nco.Float_Make (Min_relative_best_scalar_indicator_structure)
;;
let min_relative_best_scalar_indicator_rco_with_sok =
  Min_relative_best_scalar_indicator.rco_with_sok
and get_option_min_relative_best_scalar_indicator =
  Min_relative_best_scalar_indicator.get
;;

(* Min relative scalar cost decrease option. *)

module Min_relative_scalar_cost_decrease_structure =
struct
  type t = option_min_relative_scalar_cost_decrease;;

  let kind = Nonnegative;;
  let gname = "relative scalar cost decrease";;
  let name = Printf.sprintf "min %s" gname;;
  let full_name = Printf.sprintf "minimum %s" gname;;
  let default = 1e-4;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-stop-cost";;
end
;;

module Min_relative_scalar_cost_decrease =
  Nco.Float_Make (Min_relative_scalar_cost_decrease_structure)
;;
let min_relative_scalar_cost_decrease_rco_with_sok =
  Min_relative_scalar_cost_decrease.rco_with_sok
and get_option_min_relative_scalar_cost_decrease =
  Min_relative_scalar_cost_decrease.get
;;

(* Max iteration option. *)

module Max_iteration_structure =
struct
  type t = option_max_iteration;;

  let kind = Nonnegative;;
  let name = "max iteration"
  and full_name = "maximum number of iterations"
  ;;
  let default = 100;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-stop-iter";;
end
;;

module Max_iteration = Nco.Integer_Make (Max_iteration_structure);;
let max_iteration_rco_with_sok = Max_iteration.rco_with_sok
and get_option_max_iteration = Max_iteration.get
;;

(* Program configuration. *)

(* Out formatter option. *)

module Out_formatter_structure =
struct
  type t = option_out_formatter;;

  let name = "out formatter";;
  let plural_name = Printf.sprintf "%ss" name;;

  let default = Option_Std_out
  and all =
    [ Option_Std_out; Option_Std_err; Option_Out_file "<file name>" ; ]
  ;;

  let of_string = function
    | "-" | "stdout" | "Std_out" -> Option_Std_out
    | "stderr" | "Std_err" -> Option_Std_err
    | "" ->
        invalid_arg "out_formatter_of_string"
          "Empty out formatter specification"
    | s -> Option_Out_file s
  ;;

  let to_string = function
    | Option_Std_out -> "Std_out"
    | Option_Std_err -> "Std_err"
    | Option_Out_file fname -> fname
  ;;
  let opt_short_option_key = None;;
end
;;

module Out_formatter = Sco.Make (Out_formatter_structure);;
let out_formatter_rco_with_sok = Out_formatter.rco_with_sok
and get_option_out_formatter = Out_formatter.get
;;

(* Report allocated memory option. *)

let { Ar.get_flag = get_option_report_allocated_memory;
      set_flag = set_option_report_allocated_memory; _; } =
  Ar.make_flag ()
;;

let report_allocated_memory_lok =
  Ok.long_option_key_of_string "report allocated memory"
and report_allocated_memory_sok_st = Ok.Automatic
;;

let report_allocated_memory_rco =
  {
   Co.key = report_allocated_memory_lok;
   spec = Arg.Unit set_option_report_allocated_memory;
   doc =
     Printf.sprintf
       "\n\tReport total allocated memory. Default is no report.";
  }
;;

let report_allocated_memory_rco_with_sok =
  (report_allocated_memory_sok_st, report_allocated_memory_rco)
;;

(* Incremental mode option. *)

let { Ar.get_flag = get_option_incremental_mode;
      set_flag = set_option_incremental_mode; _; } =
  Ar.make_flag ()
;;

let incremental_mode_lok =
  Ok.long_option_key_of_string "incremental mode"
and incremental_mode_sok_st = Ok.Automatic
;;

let incremental_mode_rco =
  {
   Co.key = incremental_mode_lok;
   spec = Arg.Unit set_option_incremental_mode;
   doc =
     Printf.sprintf "\n\tTurn the incremental mode on. Default is off.";
  }
;;

let incremental_mode_rco_with_sok =
  (incremental_mode_sok_st, incremental_mode_rco)
;;

(* Graphics mode option. *)

module Graphics_mode_structure =
struct
  type t = option_graphics_mode;;

  let name = "graphics mode";;
  let plural_name = Printf.sprintf "%ss" name;;

  let default = Full_graphics
  and all = [ No_graphics; Parameter_graphics; Full_graphics; ]
  ;;

  let of_string = function
    | "no_graphics" | "ng" -> No_graphics
    | "parameter_graphics" | "pg" -> Parameter_graphics
    | "full_graphics" | "fg" -> Full_graphics
    | s ->
        failwith "graphics_mode_of_string"
          (Printf.sprintf "Unknown %s %S" name s)
  ;;

  let to_string = function
    | No_graphics -> "no_graphics"
    | Parameter_graphics -> "parameter_graphics"
    | Full_graphics -> "full_graphics"
  ;;
  let opt_short_option_key = None;;
end
;;

module Graphics_mode = Sco.Make (Graphics_mode_structure);;
let graphics_mode_rco_with_sok = Graphics_mode.rco_with_sok
and get_option_graphics_mode = Graphics_mode.get
;;

(* Graphics window min/max width/height options. *)

module Graphics_window_min_width_structure =
struct
  type t = option_graphics_window_min_width;;

  let kind = Positive;;
  let gname = "width";;
  let name = Printf.sprintf "graphics window min %s" gname;;
  let full_name = Printf.sprintf "graphics window minimum %s" gname;;
  let default = 800;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-gr-min-w";;
end
;;

module Graphics_window_max_width_structure =
struct
  type t = option_graphics_window_max_width;;

  let kind = Positive;;
  let gname = "width";;
  let name = Printf.sprintf "graphics window max %s" gname;;
  let full_name = Printf.sprintf "graphics window maximum %s" gname;;
  let default = 1920;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-gr-max-w";;
end
;;

module Graphics_window_min_height_structure =
struct
  type t = option_graphics_window_min_height;;

  let kind = Positive;;
  let gname = "height";;
  let name = Printf.sprintf "graphics window min %s" gname;;
  let full_name = Printf.sprintf "graphics window minimum %s" gname;;
  let default = 600;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-gr-min-h";;
end
;;

module Graphics_window_max_height_structure =
struct
  type t = option_graphics_window_max_height;;

  let kind = Positive;;
  let gname = "height";;
  let name = Printf.sprintf "graphics window max %s" gname;;
  let full_name = Printf.sprintf "graphics window maximum %s" gname;;
  let default = 1080;;
  let opt_doc = None;;
  let opt_short_option_key = Some "-gr-max-h";;
end
;;

module Graphics_window_min_width =
  Nco.Integer_Make (Graphics_window_min_width_structure)
;;
module Graphics_window_max_width =
  Nco.Integer_Make (Graphics_window_max_width_structure)
;;
module Graphics_window_min_height =
  Nco.Integer_Make (Graphics_window_min_height_structure)
;;
module Graphics_window_max_height =
  Nco.Integer_Make (Graphics_window_max_height_structure)
;;

let graphics_window_min_width_rco_with_sok =
  Graphics_window_min_width.rco_with_sok
and graphics_window_max_width_rco_with_sok =
  Graphics_window_max_width.rco_with_sok
and graphics_window_min_height_rco_with_sok =
  Graphics_window_min_height.rco_with_sok
and graphics_window_max_height_rco_with_sok =
  Graphics_window_max_height.rco_with_sok
;;

let get_option_graphics_window_min_width =
  Graphics_window_min_width.get
and get_option_graphics_window_max_width =
  Graphics_window_max_width.get
and get_option_graphics_window_min_height =
  Graphics_window_min_height.get
and get_option_graphics_window_max_height =
  Graphics_window_max_height.get
;;

(* Graphics window background color option. *)

module Graphics_window_background_color_structure =
struct
  type t = option_graphics_window_background_color;;

  let kind = Nonnegative;;
  let gname = "color";;
  let name = Printf.sprintf "graphics window bg %s" gname;;
  let full_name = Printf.sprintf "graphics window background %s" gname;;
  let default = 0xFFF5E6;; (* = x11 Seashell *)
  let opt_doc = None;;
  let opt_short_option_key = Some "-gr-bg";;
end
;;

module Graphics_window_background_color =
  Nco.Integer_Make (Graphics_window_background_color_structure)
;;
let graphics_window_background_color_rco_with_sok =
  Graphics_window_background_color.rco_with_sok
and get_option_graphics_window_background_color =
  Graphics_window_background_color.get
;;

(* Color scale names. *)

let color_scale_name_of_string name = function
  | "gray" | "grey" | "g" -> Gray
  | "cool" | "c" -> Cool
  | "hot" | "h" -> Hot
  | "jet" | "j" -> Jet
  | s ->
      failwith "color_scale_name_of_string"
        (Printf.sprintf "Unknown %s %S" name s)
;;

let color_scale_name_to_string = function
  | Gray -> "gray"
  | Cool -> "cool"
  | Hot -> "hot"
  | Jet -> "jet"
;;

let all_color_scale_names = [ Gray; Cool; Hot; Jet; ];;

(* Graphics parameter color scale name option. *)

module Graphics_parameter_color_scale_name_structure =
struct
  type t = option_graphics_color_scale_name;;

  let name = "graphics parameter color scale name";;
  let plural_name = Printf.sprintf "%ss" name;;

  let default = Hot
  and all = all_color_scale_names
  and of_string = color_scale_name_of_string name
  and to_string = color_scale_name_to_string
  and opt_short_option_key = Some "-gr-p-cs"
  ;;
end
;;

module Graphics_parameter_color_scale_name =
  Sco.Make (Graphics_parameter_color_scale_name_structure)
;;
let graphics_parameter_color_scale_name_rco_with_sok =
  Graphics_parameter_color_scale_name.rco_with_sok
and get_option_graphics_parameter_color_scale_name =
  Graphics_parameter_color_scale_name.get
;;

(* Graphics parameter min/max options. *)

module Graphics_parameter_min_structure =
struct
  type t = option_graphics_parameter_min;;

  let kind = Regular;;
  let gname = "parameter";;
  let name = Printf.sprintf "graphics %s minimum" gname;;
  let full_name = name;;
  let default = 0.;;
  let opt_doc =
    Some "Parameters under this value are clipped when displayed."
  ;;
  let opt_short_option_key = Some "-gr-p-min";;
end
;;

module Graphics_parameter_max_structure =
struct
  type t = option_graphics_parameter_max;;

  let kind = Regular;;
  let gname = "parameter";;
  let name = Printf.sprintf "graphics %s maximum" gname;;
  let full_name = name;;
  let default = 1.;;
  let opt_doc =
    Some "Parameters over this value are clipped when displayed."
  ;;
  let opt_short_option_key = Some "-gr-p-max";;
end
;;

module Graphics_parameter_min =
  Nco.Float_Make (Graphics_parameter_min_structure)
;;
module Graphics_parameter_max =
  Nco.Float_Make (Graphics_parameter_max_structure)
;;

let graphics_parameter_min_rco_with_sok = Graphics_parameter_min.rco_with_sok
and graphics_parameter_max_rco_with_sok = Graphics_parameter_max.rco_with_sok
;;

let get_option_graphics_parameter_min = Graphics_parameter_min.get
and get_option_graphics_parameter_max = Graphics_parameter_max.get
;;

(* Graphics gradient color scale name option. *)

module Graphics_gradient_color_scale_name_structure =
struct
  type t = option_graphics_color_scale_name;;

  let name = "graphics gradient color scale name";;
  let plural_name = Printf.sprintf "%ss" name;;

  let default = Gray
  and all = all_color_scale_names
  and of_string = color_scale_name_of_string name
  and to_string = color_scale_name_to_string
  and opt_short_option_key = Some "-gr-g-cs"
  ;;
end
;;

module Graphics_gradient_color_scale_name =
  Sco.Make (Graphics_gradient_color_scale_name_structure)
;;
let graphics_gradient_color_scale_name_rco_with_sok =
  Graphics_gradient_color_scale_name.rco_with_sok
and get_option_graphics_gradient_color_scale_name =
  Graphics_gradient_color_scale_name.get
;;

(* Graphics gradient amplitude option. *)

module Graphics_gradient_amplitude_structure =
struct
  type t = option_graphics_gradient_amplitude;;

  let kind = Positive;;
  let gname = "gradient amplitude";;
  let name = Printf.sprintf "graphics %s" gname;;
  let full_name = name;;
  let default = 1.;;
  let opt_doc =
    Some
      "Gradients with absolute value over this value are clipped\n\
       \twhen displayed."
  ;;
  let opt_short_option_key = Some "-gr-g-amp";;
end
;;

module Graphics_gradient_amplitude =
  Nco.Float_Make (Graphics_gradient_amplitude_structure)
;;
let graphics_gradient_amplitude_rco_with_sok =
  Graphics_gradient_amplitude.rco_with_sok
and get_option_graphics_gradient_amplitude =
  Graphics_gradient_amplitude.get
;;

(* Command-line options. *)

let phase_dimension = Co.Phase "dimension";;

let add_options () =
  Co.application_add_all_rco_with_sk
    [
     external_program_command_rco_with_sok;
     dimension_rco_with_sok;
     dimension_management_rco_with_sok;
     refinement_strategy_rco_with_sok;
     number_of_cuts_rco_with_sok;
     clear_basic_refinement_families_rco_with_sok;
     add_basic_refinement_family_rco_with_sok;
     candidate_refinements_selection_criterion_ratio_rco_with_sok;
     candidate_refinements_selection_criterion_maximum_number_rco_with_sok;
     min_relative_gradients_norm_rco_with_sok;
     min_relative_best_scalar_indicator_rco_with_sok;
     min_relative_scalar_cost_decrease_rco_with_sok;
     max_iteration_rco_with_sok;
     out_formatter_rco_with_sok;
     report_allocated_memory_rco_with_sok;
     incremental_mode_rco_with_sok;
     graphics_mode_rco_with_sok;
     graphics_window_min_width_rco_with_sok;
     graphics_window_max_width_rco_with_sok;
     graphics_window_min_height_rco_with_sok;
     graphics_window_max_height_rco_with_sok;
     graphics_window_background_color_rco_with_sok;
     graphics_parameter_color_scale_name_rco_with_sok;
     graphics_parameter_min_rco_with_sok;
     graphics_parameter_max_rco_with_sok;
     graphics_gradient_color_scale_name_rco_with_sok;
     graphics_gradient_amplitude_rco_with_sok;
    ];
  Co.phase_add phase_dimension dimension_lok;
;;

add_options ();;

let sys_argv_0 = A.get Sys.argv _0_;;

Co.set_usage_msg
  (Format.sprintf
     "\n%s -- Adaptive parameterization\n\n\
      Usage:\n\n\
      %s [<option>...] -- <cmd> [<cmd_option>...] [<cmd_argument>...]\n\n\
      Options are:@."
     sys_argv_0 sys_argv_0)
;;

(* Report/debug flags and warning/error messages. *)

let pic = Ref_indic_version.pph_id_card;;

let kps = Pic.make_option_keys pic Ok.Automatic;;

let option_key_data =
  {
   Co.name = pic.Pic.project_name;
   option_key_prefixes_opt = Some kps;
  }
;;

let printing_facilities =
  Dbg.application_printing_facilities option_key_data
;;
