(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: scalar.ml,v 1.47 2017/11/07 20:57:41 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module P = Upervasives;;
module L = Ulist;;
module L1 = Ulist1;;
module A = Uarray;;
module Dg = Udynamic_graph;;

(* Shortcuts for the Mapping library. *)

module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;

module Set = Mset;;
module Map = Mg;;

(* Shortcuts for the Refinement library. *)

module Config = Configuration;;

(* Scalar operations. *)

let qual = P.qualify "Scalar";;

(* Not imposing an explicit formatter here cause an error with output_value
   with message: "abstract value (Custom)". *)
let print_debug fname ident print =
  Config.print_debug (qual fname) ident print Format.err_formatter
;;

let debug fname = Config.debug (qual fname);;
(*let warning fname = Config.warning (qual fname);;*)

open Refinement_types;;
open Refinement_types_print;;
open Parameter;;
open Zonation;;

module Make =
  functor (Of : Uordered_field_spec.ORDERED_FIELD) ->
struct
  include Of;;

  let sum = function
    | { Map.domain = dom; images = Map.Constant y; } ->
        of_nat (Set.cardinal dom) *. y
    | { Map.images = Map.Images ims; _; } ->
        A.fold_left ( +. ) _0_ ims
  ;;

  let first_order_indicator g_zone (zt, zf) =
    let indict = sum (Map.restriction g_zone zt) in
    let indicf = sum (Map.restriction g_zone zf) in
    abs (indict -. indicf)
  ;;

  let make_overall_best_dicho g_ss dom =
    print_debug "make_overall_best_dicho" "g_ss"
      (print_gradient printer) g_ss;
    let p idx _uidx = Map.image g_ss idx < _0_ in
    Set.dichotomy p dom
  ;;

  let get_overall_best_dicho g_ss dom =
    match make_overall_best_dicho g_ss dom with
    | None -> []
    | Some dicho -> [ dicho; ]
  ;;

  let make_best_in_a_family_dichos nct bcfl1 dom =
    let bds = bounds_of_domain dom in
    L.concat
      (L1.to_list
         (L1.rev_map
            (dichotomies_from_family bds dom nct) bcfl1))
  ;;

  module Key = struct type t = Map.domain;; end;;
  module Htbl = Uhashtbl_eq.Make (Key);;

  let memo_refinements_tbl = Htbl.create 1001;;

  let memo_make_best_in_a_family_dichos nct bcfl1 dom =
    if Htbl.mem memo_refinements_tbl dom then
      Htbl.find memo_refinements_tbl dom else
    let dichos = make_best_in_a_family_dichos nct bcfl1 dom in
    Htbl.add memo_refinements_tbl dom dichos;
    dichos
  ;;

  (* Fixme: argument g of type 'a gradient should be replaced by an 'a
     partition computed using Zonation.partition_of_gradient.
     This means data structure Structural_mapping.t should be modified. *)
  let cut dm cs g _zn { Map.domain = dom; _; } =
    let ss = Set.subset_of_full g.Map.domain dom in
    let g_ss = Map.restriction g ss in
    let dichos =
      match (dm, cs) with
      | ((Vector | Best_component_only), Overall_best) ->
          get_overall_best_dicho g_ss dom
      | (Vector, Best_in_a_family (nct, bcfl1)) ->
          memo_make_best_in_a_family_dichos nct bcfl1 dom
      | (Best_component_only, Best_in_a_family (nct, bcfl1)) ->
          make_best_in_a_family_dichos nct bcfl1 dom in
    begin match dichos with
    | [] -> debug "cut" "no more possible refinements"
    | _ -> () end;
    L.map
      (fun dicho ->
        let indic = first_order_indicator g_ss dicho in
        { refinement = dicho; first_order_indicator = indic; })
      dichos
  ;;

  let initial_zonation pp1 =
    let dm = Config.get_dimension_management ()
    and cs = Config.get_refinement_strategy () in
    of_partition (cut dm cs) pp1
  ;;

  (* Vector first order indicators. *)

  let ppf = Format.err_formatter;;

  let print_refinements acr cr =
    Format.fprintf ppf "@[<hv>acr =@ %a;@ cr =@ %a@.@]"
      (print_candidate_refinement printer) acr
      (print_candidate_refinement printer) cr
  ;;

  let print_zones azone zone =
    Format.fprintf ppf "@[<hv>azone =@ %a;@ zone =@ %a@.@]"
      (print_parameter printer) azone.Dg.value
      (print_parameter printer) zone.Dg.value
  ;;

  let zero_foi cr = { cr with first_order_indicator = _0_; };;

  let zero_fois ({ Dg.image = crl; _; } as zone) =
    { zone with Dg.image = L.map zero_foi crl; }
  ;;

  let vfoi_init fz = A.map zero_fois fz;;

  let vfoi_add_foi_best_in_a_family acr cr =
    if not (same_refinement acr cr) then
      (* This function is only meant to be used with the Vector dimension
         management. In this case, all components are always subject to the
         same scalar partitions. Thus, the candidate refinements acr and cr
         must be the same. *)
      begin print_refinements acr cr; assert false end else
    let afoi = acr.first_order_indicator
    and foi = cr.first_order_indicator in
    (* Numerical tests showed that L2-norm is more efficient in this case. *)
    let afoi = afoi +. (square foi) in
    { acr with first_order_indicator = afoi; }
  ;;

  let vfoi_add_foi_overall_best acr cr =
    let afoi = acr.first_order_indicator
    and foi = cr.first_order_indicator in
    if foi <= afoi then acr else cr
  ;;

  let vfoi_add_zone vfoi_add_foi azone zone =
    if not (same_zone_geometry azone zone) then
      (* This function is only meant to be used with the Vector dimension
         management. In this case, all components are always subject to the
         same scalar partitions. Thus, the geometries of azone and zone must
         be the same. *)
      begin print_zones azone zone; assert false end else
      let acrl =
        L.map2 vfoi_add_foi
          azone.Dg.image
          zone.Dg.image in
      { azone with Dg.image = acrl; }
  ;;

  let vfoi_add_zones vfoi_add_foi = A.map2 (vfoi_add_zone vfoi_add_foi);;

  let vfoi_add_zones_overall_best =
    vfoi_add_zones vfoi_add_foi_overall_best
  and vfoi_add_zones_best_in_a_family =
    vfoi_add_zones vfoi_add_foi_best_in_a_family
  ;;
end
;;

module Float = Make (Uordered_field.Float);;
