(***********************************************************************)
(*                                                                     *)
(*                              Ref-indic                              *)
(*                                                                     *)
(*                        Refinement indicators                        *)
(*                                                                     *)
(*                       EPI Serena, INRIA Paris                       *)
(*                                                                     *)
(*  Copyright (c) 2013-2017 INRIA.                                     *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Hend Ben Ameur <hbenameur@yahoo.ca>                                *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: translate.ml,v 1.30 2017/11/07 20:57:41 fclement Exp $ *)

(* Shortcuts for the Ulib library. *)

module Nat = Unatural;;
module P = Upervasives;;
module L = Ulist;;
module L1 = Ulist1;;
module A = Uarray;;
module M = Umatrix;;
module Ddg = Udelayed_dgraph;;
module Ar = Uaref;;

let _0_ = Nat._0_;;

let array_count_occurrences p a =
  let add_occurrence accu ai = if p ai then Nat.succ accu else accu in
  A.fold_left add_occurrence _0_ a
;;

let number_of_negatives = array_count_occurrences (fun ai -> ai < 0);;

let umatrix_norm_sup abs =
  M.fold_left (fun accu mij -> max accu (abs mij)) 0.
;;

(* Shortcuts for the Basement library. *)

module A_ng = Uarray_ng;;

(* Shortcuts for the Mapping library. *)

module Num = Unumbering;;
module Midx = Umatrix_index;;
module Mnum = Umatrix_numbering;;
module Muniv = Umatrix_universe;;
module Mset = Umatrix_nonempty_set;;
module Mg = Umatrix_graph;;
module Msm = Umatrix_structural_mapping;;

module Set = Mset;;
module Map = Mg;;
module Smap = Msm;;

let zonation_of_partition z = Smap.of_partition z.Ddg.f;;

(* Shortcuts for the Refinement library. *)

module Config = Configuration;;

open Refinement_types;;
open Zonation;;

(* Translate operations. *)

let qual = P.qualify "Translate";;
let failwith = P.failwith_qual qual;;

let init_parameter_width = _0_
and init_parameter_height = _0_
;;

let { Ar.get = get_parameter_width; set = set_parameter_width; } =
  Ar.make_aref_unique "parameter width" init_parameter_width
and { Ar.get = get_parameter_height; set = set_parameter_height; } =
  Ar.make_aref_unique "parameter height" init_parameter_height
;;

let { Ar.get = get_geometry_length; set = set_geometry_length; } =
  Ar.make_aref_unique "geometry length" _0_
;;

let set_size (ni, nj) len =
  set_parameter_height ni;
  set_parameter_width nj;
  set_geometry_length len;
;;

let check_all_negative tmpa =
  let nc = A.length tmpa in
  let n = number_of_negatives tmpa in
  if n = _0_ then false else
  if n < nc then
    failwith "extract_mask"
      "mask should be identical for all components" else
  true
;;

let set_cell_masked_when_all_negative set_cell_masked idx tmpa =
  if check_all_negative tmpa then set_cell_masked (Midx.of_nat idx);
;;

let compute_mask (ni, nj as sz) mpm =
  let mask_chi = M.make ni nj P.Keep in
  let { Mnum.numbering = { Num.h_inv; _; }; _; } = Mnum.make sz in
  let set_cell_masked idx = M.set2 mask_chi (h_inv idx) P.Forget in
  let tmpm = M.transpose mpm in
  let tmpaa = M.to_array_array tmpm in
  A.iter_i (set_cell_masked_when_all_negative set_cell_masked) tmpaa;
  mask_chi
;;

let extract_mask (ni, nj as sz) mpm =
  if M.dim_j mpm <> Nat.mul ni nj then
    failwith "extract_mask"
      "column number should be the size of the domain" else
  let mask_chi = compute_mask sz mpm in
  let mask_chi_a = M.to_array mask_chi in
  let mask_chi_m =
    M.of_array_array (A.make (M.dim_i mpm) mask_chi_a) in
  let ppm =
    M.of_array_array (M.map_select Nat.of_int mask_chi_m mpm) in
  (mask_chi, ppm)
;;

let zone_add_index zn accu i ppa_i =
  if ppa_i = zn then i :: accu else accu
;;

(* mu = matrix-universe, ppa = parameter-partition-array,
   zn = zone-number, m_zn = coarse-parameter-value-in-the-zone *)
let part_add_geometry mu ppa zn m_zn =
  let rev_indices = A.fold_left_i (zone_add_index zn) [] ppa in
  let uindices = L1.of_list (L.rev_map Midx.of_nat rev_indices) in
  let dom_zn = Set.of_list1 mu uindices in
  Map.constant dom_zn m_zn
;;

(* mu = matrix-universe, ppm = parameter-partition-matrix,
   cn = component-number, ma = coarse-parameter-array *)
let add_geometry mu ppm cn ma =
  let ppa = A_ng.matrix_get_line ppm cn in
  A.map_i (part_add_geometry mu ppa) ma
;;

let part_set_geometry ppa zn { Map.domain = dom; _; } =
  Set.iter (Set.array_set ppa zn) dom
;;

let get_geometry len pp =
  let ppa = A.make len _0_ in
  A.iter_i (part_set_geometry ppa) pp;
  ppa
;;

let set_values pp ma = A.map2 Map.change_constant pp ma;;

let get_values pp = A.map Map.get_constant pp;;

let mapping_from_array { Map.domain = dom; _; } = Map.from_array dom;;

(*
open Misc_sprint;;

let sprint_matrix_universe = format_sprint Umatrix_universe_print.printer_t;;
 *)

open Optimization_status_types;;
open Stopping_criterion_types;;

module Os = Optimization_status;;

module Make =
  functor (Scal : Scalar_spec.SCALAR) ->
  functor (Comps : Components_spec.COMPONENTS) ->
struct
  module Stt = State_types.Make (Scal) (Comps);;

  type scalar = Scal.t;;
  type 'a components = 'a Comps.t;;

  type gradients = Stt.gradients;;
  type partitions = Stt.partitions;;
  type state = Stt.state;;
  type selected_refinement = Stt.selected_refinement;;

(*
  let sprint_partitions =
    format_sprint
      (Comps.printer
         (Og_printer.printer_Array
            (Umatrix_graph_print.printer_t Scal.printer)))
  ;;

  let sprint_zonations =
    format_sprint
      (Comps.printer
         (Refinement_types_print.printer_zonation Scal.printer))
  ;;

  let sprint_array_comps p =
    format_sprint (Comps.printer (Og_printer.printer_Array p))
  ;;

  let sprint_nat_array_comps = sprint_array_comps Unatural_print.printer_t
  and sprint_scal_array_comps = sprint_array_comps Scal.printer
  ;;
 *)

  let ppf = Config.get_formatter ();;

  let display_msg = Format.fprintf ppf "@[%s@]@.";;

  (* float array array -> scal array comps *)
  let sac_of_faa faa =
    Comps.of_array (A_ng.array_array_map Scal.of_float faa)
  ;;

  (* scal array comps -> float array array *)
  let sac_to_faa sas =
    A_ng.array_array_map Scal.to_float (Comps.to_array sas)
  ;;

  (* Init. *)

  let inj_init = Config.get_dimension;;

  let proj_constants j0 gm0 =
    let ng0 = umatrix_norm_sup abs_float gm0 in
    { Constants_types.initial_cost = j0;
      initial_gradients_norm = ng0; }
  ;;

  let partitions_unique_domains pps =
    let pp_0 = Comps.get pps _0_ in
    Comps.map_i (partition_unique_domains pp_0) pps
  ;;

  (* sz = size, mpm = masked-partition-matrix,
     maa = coarse-parameter-array-array *)
  let proj_coarse_init (sz, mpm, (os, maa, j)) =
    let sc =
      match os with
      | OS_Bad_initialization
      | OS_Degenerated_problem
      | OS_Stagnation
      | OS_Too_many_iterations
      | OS_Too_many_simulations ->
          display_msg
            (Printf.sprintf
               "Initial external optimization did not converge.\n\
               It returned with status: %S"
               (Os.to_string os));
          SC_No_convergent_optimizations
      | OS_Convergence -> SC_Continue in
    let (mask_chi, ppm) = extract_mask sz mpm in
(*    prerr_endline (sprint_select_matrix "mask_chi" mask_chi);*)
(*    prerr_endline (sprint_nat_matrix "ppm" ppm);*)
    let len = M.dim_j ppm in
(*    prerr_endline (sprint_nat "len" len);*)
    set_size sz len;
    let mu = Muniv.make_masked mask_chi in
(*    prerr_endline (sprint_matrix_universe "mu" mu);*)
    let mas = sac_of_faa maa in
    let pps = Comps.map_i (add_geometry mu ppm) mas in
(*    prerr_endline (sprint_partitions "pps" pps);*)
    let pps =
      match Config.get_dimension_management () with
      | Best_component_only -> pps
      | Vector -> partitions_unique_domains pps in
(*    prerr_endline (sprint_partitions "pps" pps);*)
    let zs = Comps.map Scal.initial_zonation pps in
(*    prerr_endline (sprint_zonations "zs" zs);*)
    { Stt.stopping_criterion = sc;
      zonations = zs;
      cost = j; }
  ;;

  (* Fine. *)

  let inj_fine ps =
    let pas = Comps.map Map.to_array ps in
    M.of_array_array (sac_to_faa pas)
  ;;

  let proj_fine ps gm =
    let gas = sac_of_faa (M.to_array_array gm) in
    let ngm = umatrix_norm_sup abs_float gm in
    (Comps.map2 mapping_from_array ps gas, ngm)
  ;;

  let make_fine { Stt.zonations = zs; _; } =
    let ps = Comps.map to_parameter zs in
    ((fun () -> inj_fine ps), proj_fine ps)
  ;;

  (* Coarse. *)

  let inj_coarse len pps =
(*    prerr_endline (sprint_nat "len" len);*)
(*    prerr_endline (sprint_partitions "pps" pps);*)
    let ppas = Comps.map (get_geometry len) pps in
(*    prerr_endline (sprint_nat_array_comps "ppas" ppas);*)
    let ppm = M.of_array_array (Comps.to_array ppas) in
(*    prerr_endline (sprint_nat_matrix "ppm" ppm);*)
    let mas = Comps.map get_values pps in
(*    prerr_endline (sprint_scal_array_comps "mas" mas);*)
    let maa = sac_to_faa mas in
(*    prerr_endline (sprint_float_array_array "maa" maa);*)
    (ppm, maa)
  ;;

  let proj_coarse zs pps (os, maa, j) =
    let mas = sac_of_faa maa in
    let pps = Comps.map2 set_values pps mas in
    let zs = Comps.map2 zonation_of_partition zs pps in
    { Stt.stopping_criterion = os;
      zonations = zs;
      cost = j; }
  ;;

  let make_coarse { Stt.zonations = zs; _; } pps =
    let len = get_geometry_length () in
    ((fun () -> inj_coarse len pps), proj_coarse zs pps)
  ;;
end
;;
