CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC
C                                                                     C
C                              Ref-indic                              C
C                                                                     C
C                        Refinement indicators                        C
C                                                                     C
C                       EPI Serena, INRIA Paris                       C
C                                                                     C
C  Copyright (c) 2013-2017 INRIA.                                     C
C  Institut National de Recherche en Informatique et en Automatique.  C
C  All rights reserved. This file is distributed only by permission.  C
C                                                                     C
C  Hend Ben Ameur <hbenameur@yahoo.ca>                                C
C  Francois Clement <Francois.Clement@inria.fr>                       C
C  Pierre Weis <Pierre.Weis@inria.fr>                                 C
C                                                                     C
CCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCCC

C     $Id: main.fpp,v 1.15 2017/11/06 09:05:31 fclement Exp $

#include "worker_constants.h"

      program main
C     Main program.

      implicit none

C     External routines.
      external init_engine, grad_engine, optim_engine, final_engine

      call worker_main_body
     +     (init_engine, grad_engine, optim_engine, final_engine)

      end


      subroutine init_engine(mp1, pp1, lens1, m1)
C     Initialization.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Output arguments.
      integer mp1(WORKER_MAX_MSIZE, WORKER_MAX_DIMENSION)
      integer pp1(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)
      integer lens1(WORKER_MAX_DIMENSION)
      FLOAT m1(WORKER_MAX_ZONE_NUMBER)

C     Local variables.
      character sbr*11, msg*MAX_ERROR_MSG_LENGTH
      integer s_len, deb, len, i, j

C     External function.
      logical pio_user_get_debug_mode

      s_len = 11
      sbr = 'init_engine'

      dim = 1
      deb = 0
      if (pio_user_get_debug_mode()) deb = 1

      call init_model(size_x, size_y, ndS, S, deb, ier, len, msg)
      if (ier .ne. 0) then
         call worker_error_init_engine(s_len, sbr, len, msg)
         return
      endif

      size = size_x * size_y
      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

      if (ndS .ne. size) then
         call worker_error_wrong_size(s_len, sbr, size, ndS)
         return
      endif

      zone_number = dim

C     One zone per channel.
      call worker_trivial_init(WORKER_ONE, pp1, lens1, m1)

      msize = size
      do j = 1, dim
         do i = 1, msize
            mp1(i, j) = pp1(i, j)
         enddo
      enddo

      return
      end


      subroutine grad_engine(p, g)
C     Gradient computation.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Input arguments.
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Output argument.
      FLOAT g(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character sbr*11
      integer s_len, ndT, i
      real*8 T(np_max), gdS(np_max), gdT(np_max)

      s_len = 11
      sbr = 'grad_engine'

      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

      ndT = size
      do i = 1, ndT
         T(i) = p(i, 1)
      enddo

      call gradient(ndS, ndT, S, T, gdS, gdT)

      do i = 1, ndT
         g(i, 1) = gdT(i)
      enddo

      return
      end


      subroutine optim_engine(pp, lens, m, len_sc, sc, cost)
C     Optimal coarse parameter computation.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Input arguments.
      integer pp(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)
      integer lens(WORKER_MAX_DIMENSION)

C     Input/output argument.
      FLOAT m(WORKER_MAX_ZONE_NUMBER)

C     Output argument.
      character sc*(*)
      integer len_sc
      FLOAT cost

C     Local variables.
      character sbr*12, msg*MAX_ERROR_MSG_LENGTH
      integer s_len, ncells, i, proj(np_max), nzones, j, len
      real*8 param(np_max), cout

      s_len = 12
      sbr = 'optim_engine'

      if (size .gt. np_max) then
         call worker_error_too_large_size(s_len, sbr, np_max, size)
         return
      endif

      ncells = size
      do i = 1, ncells
         proj(i) = pp(i, 1)
      enddo

      if (zone_number .gt. np_max) then
         call worker_error_too_large_size
     +        (s_len, sbr, np_max, zone_number)
         return
      endif

      if (zone_number .ne. lens(1)) then
         call worker_error_wrong_size(s_len, sbr, lens(1), zone_number)
         return
      endif

      nzones = lens(1)
      do j = 1, nzones
         param(j) = m(j)
      enddo

      call optim(ncells, proj, nzones, param, ndS, S, cout,
     +     len_sc, sc, ier, len, msg)
      if (ier .ne. 0) then
         call worker_error_optim_engine(s_len, sbr, len, msg)
         return
      endif

      do j = 1, nzones
         m(j) = param(j)
      enddo
      cost = FLOAT_OF_DOUBLE(cout)

      return
      end


      subroutine final_engine(p)
C     Finalization.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'
      include 'worker_interface_definition.fh'

C     Input argument.
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character sbr*12
      integer s_len

      s_len = 12
      sbr = 'final_engine'

      call save_opt_param(s_len, sbr, p)
      if (ier .ne. 0) return

      return
      end


      subroutine save_opt_param(s_len, sbr, p)
C     Save the optimal transmissivity parameter.

      implicit none

      include 'worker_error_definition.fh'
      include 'worker_init_definition.fh'

C     Input argument.
      character sbr*(*)
      integer s_len
      FLOAT p(WORKER_MAX_SIZE, WORKER_MAX_DIMENSION)

C     Local variables.
      character fname*13
      integer fn_len, i

      fn_len = 13
      fname = 'opt_param.dat'
      open(unit=WORKER_IOUNIT, file=fname(1:fn_len),
     +     status='unknown', access='sequential', form='formatted',
     +     iostat=ier)
      if (ier .ne. 0) then
         call worker_error_open_failure(s_len, sbr, fn_len, fname)
         return
      endif

      do i = 1, size
         write(WORKER_IOUNIT, *) p(i, 1)
      enddo

      close(WORKER_IOUNIT)

      return
      end
